package in.nosd.graylog;

import org.graylog.plugins.pipelineprocessor.EvaluationContext;
import org.graylog.plugins.pipelineprocessor.ast.expressions.Expression;
import org.graylog.plugins.pipelineprocessor.ast.functions.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.util.regex.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import javax.inject.Inject;

import com.fasterxml.jackson.databind.node.ValueNode;

import com.google.common.collect.ImmutableList;
import com.google.common.reflect.TypeToken;
import static org.graylog.plugins.pipelineprocessor.ast.functions.ParameterDescriptor.object;
import static org.graylog.plugins.pipelineprocessor.ast.functions.ParameterDescriptor.type;
import static org.graylog.plugins.pipelineprocessor.ast.functions.ParameterDescriptor.string;

public class GetListValueByRegexFunction extends AbstractFunction<List<String>> {
    Logger LOG = LoggerFactory.getLogger(Function.class);

    public static final String NAME = "get_list_val_by_regex";
    private static final String LIST_ARG = "list";
    private static final String PATTERN_ARG = "pattern";
    @SuppressWarnings("unchecked")
    private static final Class<List<String>> LIST_RETURN_TYPE = (Class<List<String>>) new TypeToken<List<String>>() {
    }.getRawType();

    private final ParameterDescriptor<Object, List<String>> listParam;
    private final ParameterDescriptor<String, String> patternParam;

    public GetListValueByRegexFunction() {
        listParam = ParameterDescriptor.object(LIST_ARG, LIST_RETURN_TYPE)
                .transform(this::transformToList)
                .description("The list of string to search regex in.")
                .build();
        patternParam = ParameterDescriptor.string(PATTERN_ARG)
                .description("The regex pattern to search in string list. Special characters needs to be double escaped.")
                .build();
    }

    private List<String> transformToList(Object value) {
        if (value instanceof Collection<?>) {
            return ((Collection<?>) value).stream()
                    .map(GetListValueByRegexFunction::convertValue)
                    // Non-text JSON nodes will return null, and type conversions are intentionally not done,
                    // since only string inputs are supported.
                    .filter(Objects::nonNull)
                    .collect(Collectors.toList());
        }
        if (value == null) {
            LOG.debug("get_list_val_by_regex: parameter \"list\" is null");
            return Collections.singletonList("");
        }
        return Collections.singletonList(value.toString());
    }

    private static String convertValue(Object o) {
        if (o instanceof ValueNode node) {
            return node.textValue();
        }
        return o.toString();
    }


    @Override
    public Object preComputeConstantArgument(FunctionArgs functionArgs, String s, Expression expression) {
        return expression.evaluateUnsafe(EvaluationContext.emptyContext());
    }

    @Override
    public List<String> evaluate(FunctionArgs functionArgs, EvaluationContext evaluationContext) {
        final List<String> list = listParam.required(functionArgs, evaluationContext);
        final String regexPattern = patternParam.required(functionArgs, evaluationContext);
        List<String> resultList = new ArrayList<String>();
        List<String> blankList = new ArrayList<String>();
        Pattern pattern;

        if (!(list instanceof List)) {
            LOG.error("get_list_val_by_regex: parameter \"list\" is of wrong type");
            return blankList;
        }

        try {
            pattern = Pattern.compile(regexPattern);
        }
        catch(PatternSyntaxException  e) {
            LOG.error("get_list_val_by_regex: invalid pattern {}", regexPattern);
            return blankList;
        }

        Matcher match;
        for (String s : list) {
            LOG.debug("get_list_val_by_regex: in loop, processing '{}'", s);
            match = pattern.matcher(s);
            if (match.matches()) {
                LOG.debug("get_list_val_by_regex match: {}", s);
                resultList.add(s);
            }
        }
        if (resultList.size() > 0) {
            return resultList;
        }

        LOG.debug("get_list_val_by_regex: No results");
        return blankList;
    }

    @Override
    public FunctionDescriptor<List<String>> descriptor() {
        return FunctionDescriptor.<List<String>>builder()
                .name(NAME)
                .description("Search values matching a regex in a string list, and return matching strings in form of a list. Match should be complete (no partial).")
                .params(ImmutableList.of(listParam, patternParam))
                .returnType(LIST_RETURN_TYPE)
                .build();
    }
}
